import abc
import hashlib
import json
import tempfile
import os

TEMP_DIR = tempfile.mkdtemp()
os.chdir(TEMP_DIR)

before = {
    "username": {
        "name": "87654321",
        "email": "username@example.org",
        "matrikel_no": "87654321",
        "submissions": [
            {
                "type": "[a01] Zeichen und Strings",
                "code": "#include <ctype.h>\nint umschalt(char *s)\n\n{\n\tint i = 0, count = 0;\n\twhile(s[i] != '\\0')\n\t{\n\t\tif(islower(s[i]) != 0)\n\t\t{\n\t\t\ts[i] = toupper(s[i]);\n\t\t} else if(isupper(s[i]) != 0)\n\t\t{\n\t\t\ts[i] = tolower(s[i]);\n\t\t} else\n\t\t{\n\t\t\tcount++;\n\t\t}\n\t\ti++;\n\t}\n\treturn count;\n}",
                "tests" : {}
            },
            {
                "type": "[a02] Iteration",
                "code": "#include <stdlib.h>\nlong iteP(unsigned int k)\n{\n\tif(k == 0)\n\t{\n\t\treturn 3;\n\t}\n\tif( k == 1)\n\t{\n\t\treturn 0;\n\t}\n\tif( k == 3)\n\t{\n\t\treturn 2;\n\t}\n\tint *p = malloc(k * sizeof(int));\n\tp[0] = 3;\n\tp[1] = 0;\n\tp[2] = 2;\n\tfor(int i = 3; i < k; i++)\n\t{\n\t\tp[i] = p[i-2] + p[i - 3];\n\t}\n\n\treturn p[k];\n}",
                "tests" : {}
            }
        ]
    }
}

after = {
  "username": {
    "name": "87654321",
    "email": "username@example.org",
    "matrikel_no": "87654321",
    "submissions": [
      {
        "type": "[a01] Zeichen und Strings",
        "code": "#include <ctype.h>\nint umschalt(char *s)\n\n{\n\tint i = 0, count = 0;\n\twhile(s[i] != '\\0')\n\t{\n\t\tif(islower(s[i]) != 0)\n\t\t{\n\t\t\ts[i] = toupper(s[i]);\n\t\t} else if(isupper(s[i]) != 0)\n\t\t{\n\t\t\ts[i] = tolower(s[i]);\n\t\t} else\n\t\t{\n\t\t\tcount++;\n\t\t}\n\t\ti++;\n\t}\n\treturn count;\n}",
        "tests": {
          "EmptyTest": {
            "name": "EmptyTest",
            "annotation": "",
            "label": "NOT_EMPTY"
          }
        }
      },
      {
        "type": "[a02] Iteration",
        "code": "#include <stdlib.h>\nlong iteP(unsigned int k)\n{\n\tif(k == 0)\n\t{\n\t\treturn 3;\n\t}\n\tif( k == 1)\n\t{\n\t\treturn 0;\n\t}\n\tif( k == 3)\n\t{\n\t\treturn 2;\n\t}\n\tint *p = malloc(k * sizeof(int));\n\tp[0] = 3;\n\tp[1] = 0;\n\tp[2] = 2;\n\tfor(int i = 3; i < k; i++)\n\t{\n\t\tp[i] = p[i-2] + p[i - 3];\n\t}\n\n\treturn p[k];\n}",
        "tests": {}
      }
    ]
  }
}


def all_subclasses(cls):
    return cls.__subclasses__() \
        + [g for s in cls.__subclasses__() for g in all_subclasses(s)]

def sha1_prefix(submission_obj):
    return hashlib.sha1(
        submission_obj['code'].encode()
    ).hexdigest()[:12]

class Test(metaclass=abc.ABCMeta):
    """docstring for IliasQuestion"""

    @classmethod
    def available_tests(cls):
        return {sub.__name__ : sub for sub in all_subclasses(cls)}

    def __new__(cls, *args, **kwargs):
        assert hasattr(cls, 'depends'),         "depends not defined"
        assert hasattr(cls, 'label_success'),   "label_success not defined"
        assert hasattr(cls, 'label_failure'),   "label_failure not defined"
        return super().__new__(cls)

    def __init__(self, submission_obj):

        if not self.dependencies_satisfied(submission_obj):
            self.result     = self.label_failure
            self.annotation = "TEST DEPENDENCY NOT MET"
            return

        elif str(self) in submission_obj['tests']:
            self.deserialize(submission_obj['tests'][str(self)])

        else:
            self.result, self.annotation = self.run_test(submission_obj)
            self.serialize(submission_obj)

    def __bool__(self):
        return self.result

    def __str__(self):
        return self.__class__.__name__

    def dependencies_satisfied(self, submission_obj):
        return all(dep(submission_obj).result for dep in self.depends)

    def deserialize(self, test):
        self.result     = test['label'] == self.label_success
        self.annotation = test['annotation']

    def serialize(self, submission_obj):
        as_dict = {
            'name'       : str(self),
            'annotation' : self.annotation
        }

        if self.result:
            as_dict['label'] = self.label_success
        else:
            as_dict['label'] = self.label_failure

        submission_obj['tests'][str(self)] = as_dict

    @abc.abstractmethod
    def run_test(self, submission_obj) -> (bool, str):
        return NotImplemented


class EmptyTest(Test):
    """docstring for EmptyTest"""

    depends         = ()
    label_success   = 'NOT_EMPTY'
    label_failure   = 'EMPTY'

    def run_test(self, submission_obj):
        return bool(submission_obj['code'].strip()), ""


class CompileTest(Test):

    depends         = (EmptyTest, )
    label_success   = 'COMPILATION_SUCCESSFUL'
    label_failure   = 'COMPILATION_FAILED'

    def run_test(self, submission_obj):
        prefix = sha1_prefix(submission_obj)
        os.mkdir(prefix)

        return False, "naa"





d = CompileTest(before["username"]['submissions'][0])


with open("/dev/null", "w") as out:
    out.write(json.JSONEncoder().encode(after))
